import sys 

class TextWrappers:
    def __init__(self):
        if (sys.platform == 'win32'):
            self.LOCK = ''
            self.ERROR = ''
            self.SUCCESS = ''
            self.INFO = ''
            self.BANNER = ''
            self.PARAM = ''
            self.DEFAULT = ''
            self.HEADER = ''
            self.DEFAULT = ''
        else: # MacOS, Linux, etc.
            self.LOCK = '\U0001F510 '
            self.ERROR = '\033[91m\u2717 '
            self.SUCCESS = '\033[92m\u2713 '
            self.INFO = '\033[96m'
            self.BANNER = '\033[1m'
            self.PARAM = '\x1b[3m'
            self.DEFAULT = '\033[0m\x1b[0m'
            self.HEADER = '\033[94m\033[1m'
            self.DEFAULT = '\033[0m\x1b[0m'


import os
import sqlite3 

from cryptography.fernet import Fernet 
from getpass import getpass
import pyperclip

from .text_wrappers import TextWrappers


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


class PasswordSafe:
    def __init__(self, key):
        self._connection = sqlite3.connect(ROOT_DIR + '/password_safe.db')
        self._cursor = self._connection.cursor()

        self._cursor.execute('CREATE TABLE IF NOT EXISTS entries (title text NOT NULL PRIMARY KEY, category text NOT NULL, username text NOT NULL, password blob NOT NULL)')
        
        self._key = key

        self._text = TextWrappers()


    def add(self, entry_name):
        if self._exists(entry_name):
            print(self._text.ERROR + "Entry '" + entry_name + "' already exists." + self._text.DEFAULT)
            return 
        
        category = input("Group (e.g. personal, work, etc.): ").upper()
        username = input("Username: ")
        password = getpass(prompt = "Password: ")

        encrypted_password = self._encrypt(password)

        params = (entry_name, category, username, encrypted_password)
        query = 'INSERT INTO entries (title, category, username, password) VALUES (?, ?, ?, ?)'
        self._cursor.execute(query, params)
        self._connection.commit()

        print(self._text.SUCCESS + "Success! Entry stored." + self._text.DEFAULT)


    def delete(self, entry_name):
        if not self._exists(entry_name):
            print(self._text.ERROR + "Entry '" + entry_name + "' not found." + self._text.DEFAULT)
            return 

        confirmation = input("Are you sure you want to delete '" + entry_name + "'? [Y/n] ").strip().lower()
        if confirmation == 'y' or confirmation == 'yes':
            params = (entry_name, )
            query = 'DELETE FROM entries WHERE title = (?)'
            self._cursor.execute(query, params)
            self._connection.commit() 

            print(self._text.SUCCESS + "Success! Entry deleted." + self._text.DEFAULT)

    def edit(self, entry_name):
        if not self._exists(entry_name):
            print(self._text.ERROR + "Entry '" + entry_name + "' not found." + self._text.DEFAULT)
            return 

        params = (entry_name, )
        query  = 'SELECT category, username, password FROM entries WHERE title = (?)'
        self._cursor.execute(query, params)
        entry = self._cursor.fetchone()

        curr_title = entry_name
        curr_category = entry[0] 
        curr_username = entry[1] 
        curr_password = self._decrypt(entry[2])

        title = input("Title [" + curr_title + "]: ")
        category = input("Group [" + curr_category + "]: ").upper()
        username = input("Username [" + curr_username + "]: ")
        password = getpass(prompt = "Password [" + curr_password + "]: ")

        if title.strip() == '':
            title = curr_title
        if category.strip() == '':
            category = curr_category
        if username.strip() == '':
            username = curr_username
        if password == '':
            password = curr_password

        password = self._encrypt(password)

        params = (title, category, username, password, entry_name)
        query = 'UPDATE entries SET title=(?), category=(?), username=(?), password=(?) WHERE title = (?)'
        self._cursor.execute(query, params)
        self._connection.commit() 

        print(self._text.SUCCESS + "Success! Entry updated." + self._text.DEFAULT)


    def peek(self, entry_name):
        if not self._exists(entry_name):
            print(self._text.ERROR + "Entry '" + entry_name + "' not found." + self._text.DEFAULT)
            return 

        params = (entry_name, )
        query = 'SELECT username, password FROM entries WHERE title = (?)'
        self._cursor.execute(query, params)
        entry = self._cursor.fetchone()

        decrypted_password = self._decrypt(entry[1])

        print("Username ==> " + entry[0])
        print("Password ==> " + decrypted_password)


    def copy(self, entry_name):
        if not self._exists(entry_name):
            print(self._text.ERROR + "Entry '" + entry_name + "' not found." + self._text.DEFAULT)
            return

        params = (entry_name, )
        query = 'SELECT password FROM entries WHERE title = (?)'
        self._cursor.execute(query, params)
        entry = self._cursor.fetchone() 

        decrypted_password = self._decrypt(entry[0])

        pyperclip.copy(decrypted_password)


    def list_all(self):
        self._cursor.execute('SELECT category, title FROM entries')
        entries = self._cursor.fetchall()

        entries_count = len(entries)

        print("Password Safe (" + str(entries_count) + " entries)")

        if entries_count == 0:
            return 

        database = {}
        for category, entry in entries: 
            database.setdefault(category, []).append(entry) 

        self._display_tree(database)


    def close(self):
        self._connection.close()


    def _encrypt(self, password):
        f = Fernet(self._key) 

        password = bytes(password, 'ascii')
        token = f.encrypt(password)

        return token


    def _decrypt(self, token):
        f = Fernet(self._key)

        password = f.decrypt(token)
        password = password.decode('utf-8')

        return password


    def _exists(self, entry_name):
        params = (entry_name, )
        query = 'SELECT title FROM entries WHERE title = (?)'
        self._cursor.execute(query, params)
        entry = self._cursor.fetchone()

        if entry != None:
            return True 

        return False


    def _display_tree(self, database):
        branch = '├─'
        pipe = '│'
        end = '└─'
        dash = '── '
        indent = ' ' * 5
        tab = ' ' * 4

        last_category = list(database.keys())[len(database.keys()) - 1]
        is_last_category = False

        for category in database.keys():
            if category == last_category:
                print(end + dash + self._text.HEADER + category + self._text.DEFAULT)
                is_last_category = True
            else:
                print(branch +  dash + self._text.HEADER + category + self._text.DEFAULT)

            last_entry = (database[category])[len(database[category]) - 1]
            for entry in database[category]:
                if entry == last_entry and is_last_category:
                    print(indent + end + dash + entry)
                elif entry == last_entry and not is_last_category:
                    print(pipe + tab + end + dash + entry)
                elif is_last_category:
                    print(indent + branch + dash + entry)
                else: 
                    print(pipe + tab + branch + dash + entry)


import os
import sys

import base64
import bcrypt
from getpass import getpass
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes 
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from .password_safe import PasswordSafe
from .text_wrappers import TextWrappers


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


class PasswordManager:
    def __init__(self):
        self._password_safe = None
        self._text = TextWrappers()

    def _destroy_safe(self):
        confirmation = input("Are you sure you want to destroy this safe? All data including the master PIN will be deleted. [Y/n] ").strip().lower()

        self._password_safe.close()

        if confirmation == 'y' or confirmation == 'yes':
            os.remove(ROOT_DIR + '/master.key')
            os.remove(ROOT_DIR + '/password_safe.db')

            print(self._text.SUCCESS + "Success! Password safe destroyed. Restart the program to set up a new safe." + self._text.DEFAULT)

            return True
        
        return False


    def _display_options(self):
        print("add <" + self._text.PARAM + "entry" + self._text.DEFAULT + ">        - Create and store an entry into the safe.")
        print("delete <" + self._text.PARAM + "entry" + self._text.DEFAULT + ">     - Remove an entry from the safe.")
        print("edit <" +  self._text.PARAM + "entry" + self._text.DEFAULT + ">       - Change and update an entry's information.")
        print("peek <" + self._text.PARAM + "entry" + self._text.DEFAULT + ">       - Display the username and password of an entry.")
        print("copy <" + self._text.PARAM + "entry" + self._text.DEFAULT + ">       - Copy the password of an entry without peeking.")
        print("list               - Display all the entries stored in the safe.")
        print("help               - Display available options and their usage details.")
        print("destroy            - Delete all data including the master PIN.")
        print("exit               - Lock the safe and exit the program.")


    def _manage_actions(self):
        print(self._text.INFO + "\nUse the command 'help' for usage details\nUse the command 'exit' to exit at any time.\n" + self._text.DEFAULT)

        cmd = ''
        
        while True:
            cmd = input(">> ").split(maxsplit = 1)

            if len(cmd) == 1:
                if cmd[0].lower() == 'help':
                    self._display_options()
                elif cmd[0].lower() == 'list':
                    self._password_safe.list_all()
                elif cmd[0].lower() == 'destroy':
                    if self._destroy_safe():
                        return
                elif cmd[0].lower() == 'exit':
                    self._password_safe.close()
                    return
                elif cmd[0].lower() in ('add', 'delete', 'edit', 'peek', 'copy'):
                    print(self._text.ERROR + "Command '" + cmd[0].lower() + "' requires an additional parameter. Type 'help' for usage details." + self._text.DEFAULT)
                else:
                    print(self._text.ERROR + "Command '" + cmd[0] + "' not found." + self._text.DEFAULT)

            if len(cmd) == 2:
                if cmd[0].lower() == 'add':
                    self._password_safe.add(cmd[1])
                elif cmd[0].lower() == 'delete':
                    self._password_safe.delete(cmd[1])
                elif cmd[0].lower() == 'edit':
                    self._password_safe.edit(cmd[1])
                elif cmd[0].lower() == 'peek':
                    self._password_safe.peek(cmd[1])
                elif cmd[0].lower() == 'copy':
                    self._password_safe.copy(cmd[1])
                else:
                    print(self._text.ERROR + "Command '" + cmd[0] + "' not found." + self._text.DEFAULT)


    def _derive_key(self, pin):
        salt = b'V\x04<\x7f\x07\x89\xb1\xe5\x8dF\x02\xb2\x85\xb6\x9fw'

        kdf = PBKDF2HMAC(
            algorithm = hashes.SHA256(),
            length = 32,
            salt = salt,
            iterations = 100000,
            backend = default_backend()
        )

        key = base64.urlsafe_b64encode(kdf.derive(pin))

        return key


    def _auth_user(self):
        try:
            f = open(ROOT_DIR + '/master.key', 'br')
            hashed_pin = f.read()
        except FileNotFoundError:
            print(self._text.ERROR + "User data is corrupted or lost. Unable to authroize user." + self._text.DEFAULT) 

        while True:
            pin = getpass(prompt = "Enter your master PIN: ")
            pin = bytes(pin, 'ascii')

            if bcrypt.checkpw(pin, hashed_pin):
                key = self._derive_key(pin)
                self._password_safe = PasswordSafe(key)

                print(self._text.SUCCESS + "Success! Password safe unlocked." + self._text.DEFAULT)
                return
            else:
                print(self._text.ERROR + "Incorrect PIN. Please try again." + self._text.DEFAULT)


    def _validate_pin(self, pin, pin_confirmation):
        # validate that input is not blank
        if pin == '':
            print(self._text.ERROR + "No PIN was entered. Please try again." + self._text.DEFAULT)
            return False 

        # validate that pin is a combination of digits 0-9
        try:
            int(pin)
        except ValueError:
            print(self._text.ERROR + "PIN was not a combination of digits 0-9. Please try again." + self._text.DEFAULT)
            return False 

        # validate that pin is 4-digits long
        if len(pin) != 4:
            print(self._text.ERROR + "PIN must be 4-digits long. Please try again." + self._text.DEFAULT)
            return False 

        # validate that pin was confirmed
        if pin != pin_confirmation:
            print(self._text.ERROR + "The PIN numbers entered did not match. Please try again." + self._text.DEFAULT)
            return False

        return True
        
            
    def _create_user(self):
        print(self._text.INFO + "Welcome! To get started, create a master PIN. Do not lose it as it is unrecoverable." + self._text.DEFAULT)

        while True: 
            pin = getpass(prompt = "Enter a 4-digit PIN: ") 
            pin_confirmation = getpass(prompt = "Confirm your PIN: ")

            if self._validate_pin(pin, pin_confirmation):
                pin = bytes(pin, 'ascii')
                hashed_pin = bcrypt.hashpw(pin, bcrypt.gensalt())

                f = open(ROOT_DIR + '/master.key', 'wb')
                f.write(hashed_pin)
                f.close()

                key = self._derive_key(pin)
                self._password_safe = PasswordSafe(key)

                print(self._text.SUCCESS + "Success! Password safe configured. Restart the program and enter your master PIN to begin." + self._text.DEFAULT)
                
                return


    def _is_new_user(self):
        if os.path.exists(ROOT_DIR + '/master.key') and os.path.exists(ROOT_DIR + '/password_safe.db'):
            return False
        return True


    def _display_banner(self):
        print("\n===========================================================")
        print("\t" + self._text.BANNER + self._text.LOCK + "PYTHON PASSWORD SAFE - Command Line Tool" + self._text.DEFAULT)
        print("===========================================================\n")
    

    def run(self):
        self._display_banner()

        if self._is_new_user():
            self._create_user() 
        else:
            self._auth_user()
            self._manage_actions()

def main():
    try: 
        PasswordManager().run()
    except KeyboardInterrupt:
        print("\n")
        sys.exit(0)

if __name__ == "__main__":
    main()


import setuptools 

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup (
    name = 'py-pwsafe',
    version = '2.0.1',
    description = 'Python Password Safe',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = 'Ricky Thai',
    author_email = 'rickyvanthai@gmail.com',
    url = 'https://github.com/rvthai/python-passwordsafe',
    package_dir={'pypwsafe': 'src'},
    packages = ['pypwsafe'],
    install_requires=['bcrypt', 'pyperclip', 'cryptography'],  
    entry_points = {
        'console_scripts': ['pypwsafe = pypwsafe.password_manager:main'],
    },
    license = 'MIT',
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires = '>=3.6'
)